<?php

use Monolog\ErrorHandler;
use Monolog\Handler\StreamHandler;
use Monolog\Logger;
use php\Database;
use php\Mailer;
use php\Mediawiki;
use php\Response;
use php\TrackingManager;
use php\UserManager;

/** @noinspection PhpIncludeInspection Exists after `npm run deploy` */
require_once __DIR__ . "/vendor/autoload.php";


/// Preparations
// Load config
// TODO: Check permissions, exit if too permissive
$config = parse_ini_file("config.default.ini.php", process_sections: true, scanner_mode: INI_SCANNER_TYPED) or exit(1);
if (file_exists("config.ini.php")) {
    $config_custom =
        parse_ini_file("config.ini.php", process_sections: true, scanner_mode: INI_SCANNER_TYPED) or exit(1);
    $config = array_merge($config, $config_custom);
}

// Create logger
$logger = new Logger("main");
$logger->pushHandler(new StreamHandler($config["logger"]["file"], $config["logger"]["level"]));
ErrorHandler::register($logger);

// Connect to database
$db_exists = file_exists($config["database"]["filename"]);
$conn = Database::connect($config["database"]["filename"]);

// Instantiate utility classes
$mediawiki = new Mediawiki($logger->withName("Mediawiki"));
$user_manager = new UserManager($conn);
$tracking_manager = new TrackingManager($mediawiki, $conn);
$mailer = new Mailer($logger->withName("Mailer"), $config, $conn);

// Create db if it does not exist
if (!$db_exists) {
    $logger->warning("Database does not exist. Creating new database at '{$config["database"]["filename"]}'.");

    $user_manager->install();
    $tracking_manager->install();
    $mailer->install();
}

// Start session
/**
 * Generates a CSRF token.
 *
 * @return string the generated CSRF token
 */
function generate_csrf_token(Logger $logger): string
{
    try {
        return bin2hex(random_bytes(32));
    } catch (Exception $exception) {
        $logger->emergency("Failed to generate token.", ["cause" => $exception]);
        http_response_code(500);
        exit();
    }
}

session_start();
if (!isset($_SESSION["token"]))
    $_SESSION["token"] = generate_csrf_token($logger);

// Read JSON from POST
$post_input = file_get_contents("php://input");
if ($post_input !== false) $_POST = json_decode($post_input, associative: true);


/// Define validation helpers
/**
 * Validates the user's CSRF token by comparing `$_POST["token"]` against `$_SESSION["token"]`.
 *
 * @return Response|null `null` if the user's CSRF token is valid, or an error response if either value is not set or if
 * the two are not equal
 */
function validate_csrf(): ?Response
{
    if (!(isset($_POST["token"]) && isset($_SESSION["token"]) && hash_equals($_POST["token"], $_SESSION["token"])))
        return new Response(
            payload: ["target" => null, "message" => "Invalid CSRF token. Try refreshing the page."],
            satisfied: false
        );

    return null;
}

/**
 * Validates that the user is logged out by checking that `$_SESSION["uuid"]` is not set.
 *
 * @return Response|null `null` if the user is logged out, or an error response otherwise
 */
function validate_logged_out(): ?Response
{
    if (isset($_SESSION["uuid"]))
        return new Response(payload: ["target" => null, "message" => "User is already logged in."], satisfied: false);

    return null;
}

/**
 * Validates that the user is logged in by checking that `$_SESSION["uuid"]` is set.
 *
 * @return Response|null `null` if the user is logged in, or an error response otherwise
 */
function validate_logged_in(): ?Response
{
    if (!isset($_SESSION["uuid"]))
        return new Response(payload: ["target" => null, "message" => "User is not logged in."], satisfied: false);

    return null;
}

/**
 * Validates that each of `arguments` is set.
 *
 * @param mixed ...$arguments the arguments to check
 * @return Response|null `null` if all `arguments` are set, or an error response otherwise
 */
function validate_has_arguments(mixed ...$arguments): ?Response
{
    foreach (func_get_args() as $argument) {
        if (!isset($argument))
            return new Response(payload: ["target" => null, "message" => "Missing argument."], satisfied: false);
        if (!is_string($argument))
            return new Response(payload: ["target" => null, "message" => "Invalid argument type."], satisfied: false);
    }

    return null;
}


/// Process request
$response = null;

if (isset($_POST["action"])) {
    // POST requests; alter state
    switch ($_POST["action"]) {
        case "register":
            $response = validate_csrf()
                ?? validate_logged_out()
                ?? validate_has_arguments($_POST["email"], $_POST["password"], $_POST["password_confirm"])
                ?? $user_manager->register($mailer, $_POST["email"], $_POST["password"], $_POST["password_confirm"]);
            break;
        case "login":
            $response = validate_csrf()
                ?? validate_logged_out()
                ?? validate_has_arguments($_POST["email"], $_POST["password"]);
            if ($response !== null) break;

            [$response, $uuid] = $user_manager->check_login($_POST["email"], $_POST["password"]);
            if ($response->satisfied) $_SESSION["uuid"] = $uuid;
            break;
        case "logout":
            $response = validate_csrf() ?? validate_logged_in();
            if ($response !== null) break;

            session_destroy();
            session_start();
            $_SESSION["token"] = generate_csrf_token($logger);
            $response = new Response(payload: null, satisfied: true);
            break;
        case "update-email":
            $response = validate_csrf()
                ?? validate_logged_in()
                ?? validate_has_arguments($_POST["email"])
                ?? $user_manager->set_email($_SESSION["uuid"], $_POST["email"], $mailer);
            break;
        case "verify-email":
            $response = validate_has_arguments($_POST["email"], $_POST["token"])
                ?? $user_manager->verify_email($_POST["email"], $_POST["token"]);
            break;
        case "resend-verify-email":
            $response = validate_csrf()
                ?? validate_logged_in()
                ?? $user_manager->resend_verify_email($_SESSION["uuid"], $mailer);
            break;
        case "update-password":
            $response = validate_csrf()
                ?? validate_logged_in()
                ?? validate_has_arguments($_POST["password_old"], $_POST["password_new"], $_POST["password_confirm"])
                ?? $user_manager->set_password(
                    $_SESSION["uuid"],
                    $_POST["password_old"],
                    $_POST["password_new"],
                    $_POST["password_confirm"]
                );
            break;
        case "user-delete":
            $response = validate_csrf()
                ?? validate_logged_in()
                ?? $user_manager->delete($_SESSION["uuid"]);
            break;
        case "add-tracking":
            $response = validate_csrf()
                ?? validate_logged_in()
                ?? validate_has_arguments($_POST["person_name"])
                ?? $tracking_manager->add_tracking($_SESSION["uuid"], $_POST["person_name"]);
            break;
        case "remove-tracking":
            $response = validate_csrf()
                ?? validate_logged_in()
                ?? validate_has_arguments($_POST["person_name"])
                ?? $tracking_manager->remove_tracking($_SESSION["uuid"], $_POST["person_name"]);
            break;
        default:
            $response = new Response(
                payload: ["target" => null, "message" => "Unknown POST action '{$_POST["action"]}'."],
                satisfied: false
            );
            break;
    }
} elseif (isset($_GET["action"])) {
    // GET requests; do not alter state
    switch ($_GET["action"]) {
        case "start-session":
            if (!isset($_SESSION["uuid"])) {
                $response = new Response(payload: ["target" => null, "message" => null], satisfied: false);
                break;
            }

            $response = $user_manager->user_exists($_SESSION["uuid"]);
            if (!$response->satisfied) {
                session_destroy();
                session_start();
                $_SESSION["token"] = generate_csrf_token($logger);
            }
            break;
        case "get-user-data":
            $response = validate_logged_in() ?? $user_manager->get_user_data($_SESSION["uuid"]);
            break;
        case "list-trackings":
            $response = validate_logged_in() ?? $tracking_manager->list_trackings($_SESSION["uuid"]);
            break;
        default:
            $response = new Response(
                payload: ["target" => null, "message" => "Unknown GET action '{$_GET["action"]}'."],
                satisfied: false
            );
    }
} elseif ($argc > 1) {
    // CLI
    if (hash_equals($config["admin"]["cli_secret"], "REPLACE THIS WITH A SECRET VALUE"))
        exit("Default value for 'cli_secret' detected. Feature disabled.");
    if (!hash_equals($config["admin"]["cli_secret"], $argv[2]))
        exit("Incorrect value for 'cli_secret'.");

    switch ($argv[1]) {
        case "update-all-trackings":
            $logger->info("Updating all trackings.");
            $tracking_manager->update_trackings($mailer, $tracking_manager->list_all_unique_person_names());
            exit("Successfully updated all trackings.");
        case "process-email-queue":
            $logger->info("Processing email queue.");
            $mailer->process_queue();
            exit("Successfully processed email queue.");
        default:
            exit("Unknown CLI action '$argv[1]'.");
    }
} else {
    // No action given, nothing done, so that's a success
    $response = new Response(payload: null, satisfied: true);
}

header("Content-type:application/json;charset=utf-8");
exit(json_encode(array(
    "payload" => $response->payload,
    "satisfied" => $response->satisfied,
    "token" => $_SESSION["token"]
)));
