<?php

namespace php;

use DateTime;
use PDO;


/**
 * Manages interaction with the database in the context of users.
 */
// TODO: Remove duplication in this class
class UserManager
{
    /**
     * The maximum length of an email address.
     */
    private const MAX_EMAIL_LENGTH = 254;
    /**
     * The minimum length of a password;
     */
    private const MIN_PASSWORD_LENGTH = 8;
    /**
     * The maximum length of a password.
     */
    private const MAX_PASSWORD_LENGTH = 64;
    /**
     * The minimum number of minutes between two verification emails.
     */
    private const MINUTES_BETWEEN_VERIFICATION_EMAILS = 5;

    /**
     * @var PDO The database connection to interact with.
     */
    private PDO $conn;


    /**
     * Constructs a new user manager.
     *
     * @param PDO $conn the database connection to interact with
     */
    public function __construct(PDO $conn)
    {
        $this->conn = $conn;
    }


    /**
     * Populates the database with the necessary structures for users.
     *
     * @return void
     */
    public function install(): void
    {
        $this->conn->exec("CREATE TABLE users(uuid text not null unique primary key default(lower(hex(randomblob(16)))),
                                              email text not null unique,
                                              email_verification_token text,
                                              email_verification_token_timestamp int not null,
                                              password text not null,
                                              password_update_time int not null);");
    }

    /**
     * Registers a new user.
     *
     * @param Mailer $mailer the mailer to notify the user with
     * @param string $email the user-submitted email address
     * @param string $password the user-submitted password
     * @param string $password_confirm the user-submitted password confirmation
     * @return Response a response with message `null` if the user was registered, or a response with a message
     * explaining what went wrong otherwise
     */
    public function register(Mailer $mailer, string $email, string $password, string $password_confirm): Response
    {
        // Validate
        if (!filter_var($email, FILTER_VALIDATE_EMAIL) || strlen($email) > self::MAX_EMAIL_LENGTH)
            return new Response(
                payload: ["target" => "email", "message" => "Invalid email address."],
                satisfied: false
            );
        if (strlen($password) < self::MIN_PASSWORD_LENGTH || strlen($password) > self::MAX_PASSWORD_LENGTH)
            return new Response(
                payload: [
                    "target" => "password",
                    "message" => "Your password should be at least 8 and at most 64 characters long."
                ],
                satisfied: false
            );
        if ($password !== $password_confirm)
            return new Response(
                payload: ["target" => "passwordConfirm", "message" => "Passwords do not match."],
                satisfied: false
            );

        // Begin transaction
        $this->conn->beginTransaction();

        // Check if email address is already in use
        $stmt = $this->conn->prepare("SELECT COUNT(*) AS count
                                      FROM users
                                      WHERE email=:email;");
        $stmt->bindValue(":email", $email);
        $stmt->execute();
        $result = $stmt->fetch(PDO::FETCH_ASSOC);
        if ($result["count"] > 0) {
            $this->conn->rollBack();
            return new Response(
                payload: ["target" => "email", "message" => "Email address already in use."],
                satisfied: false
            );
        }

        // Register user
        $stmt = $this->conn->prepare("INSERT INTO users (email,
                                                         email_verification_token,
                                                         email_verification_token_timestamp,
                                                         password,
                                                         password_update_time)
                                      VALUES (:email,
                                              lower(hex(randomblob(16))),
                                              unixepoch(),
                                              :password,
                                              unixepoch())
                                      RETURNING email_verification_token;");
        $stmt->bindValue(":email", $email);
        $stmt->bindValue(":password", password_hash($password, PASSWORD_DEFAULT));
        $stmt->execute();
        $email_verification_token = $stmt->fetchAll(PDO::FETCH_ASSOC)[0]["email_verification_token"];

        // Respond
        $this->conn->commit();
        $mailer->queue_registration($email, $email_verification_token);
        return new Response(payload: null, satisfied: true);
    }

    /**
     * Deletes the user with the given UUID.
     *
     * @param string $uuid the UUID of the user to delete
     * @return Response a response with message `null` if the user was deleted, or a response with a message explaining
     * what went wrong otherwise
     */
    public function delete(string $uuid): Response
    {
        $stmt = $this->conn->prepare("DELETE FROM users WHERE uuid=:uuid;");
        $stmt->bindValue(":uuid", $uuid);
        $stmt->execute();

        $stmt = $this->conn->prepare("DELETE FROM trackings WHERE user_uuid=:uuid");
        $stmt->bindValue(":uuid", $uuid);
        $stmt->execute();

        return new Response(payload: null, satisfied: true);
    }

    /**
     * Validates a login attempt with the given email address and password.
     *
     * @param string $email the email address of the user whose password should be checked
     * @param string $password the password to check against the specified user
     * @return array{0: Response, 1: string|null} a response and the user's UUID
     */
    public function check_login(string $email, string $password): array
    {
        if (!filter_var($email, FILTER_VALIDATE_EMAIL) || strlen($email) > self::MAX_EMAIL_LENGTH)
            return [
                new Response(
                    payload: ["target" => "email", "message" => "Invalid email address."],
                    satisfied: false
                ),
                null
            ];
        if (strlen($password) > self::MAX_PASSWORD_LENGTH)
            return [
                new Response(
                    payload: ["target" => "password", "message" => "Incorrect combination of email and password."],
                    satisfied: false
                ),
                null
            ];

        $stmt = $this->conn->prepare("SELECT uuid, password FROM users WHERE email=:email;");
        $stmt->bindValue(":email", $email);
        $stmt->execute();
        $results = $stmt->fetchAll(PDO::FETCH_ASSOC);

        if (sizeof($results) === 0 || !password_verify($password, $results[0]["password"])) {
            return [
                new Response(
                    payload: ["target" => "password", "message" => "Incorrect combination of email and password."],
                    satisfied: false
                ),
                null
            ];
        }

        return [new Response(payload: null, satisfied: true), $results[0]["uuid"]];
    }

    /**
     * Returns a satisfied response if a user with the given UUID exists, or an unsatisfied response otherwise.
     *
     * @param string $uuid the UUID of the user to check
     * @return Response a satisfied response if a user with the given UUID exists, or an unsatisfied response otherwise
     */
    public function user_exists(string $uuid): Response
    {
        $stmt = $this->conn->prepare("SELECT EXISTS(SELECT 1 FROM users WHERE uuid=:uuid);");
        $stmt->bindValue(":uuid", $uuid);
        $stmt->execute();
        $result = $stmt->fetch();

        return $result[0] === 1
            ? new Response(payload: null, satisfied: true)
            : new Response(payload: ["target" => null, "message" => null], satisfied: false);
    }

    /**
     * Returns the user with the given UUID.
     *
     * @param string $uuid the UUID of the user to return
     * @return Response the user with the given UUID, or a response with an explanation what went wrong otherwise
     */
    public function get_user_data(string $uuid): Response
    {
        $stmt = $this->conn->prepare("SELECT uuid, email, email_verification_token IS NULL AS email_is_verified,
                                             password_update_time
                                      FROM users
                                      WHERE uuid=:uuid;");
        $stmt->bindValue(":uuid", $uuid);
        $stmt->execute();
        $user = $stmt->fetch(PDO::FETCH_ASSOC);
        if ($user === false)
            return new Response(
                payload: ["target" => "uuid", "message" => "Something went wrong. Please try logging in again."],
                satisfied: false
            );

        return new Response(payload: $user, satisfied: true);
    }

    /**
     * Updates the indicated user's email address.
     *
     * @param string $uuid the UUID of the user whose email address should be updated
     * @param string $email the new email address
     * @param Mailer $mailer the mailer to send a notification with to the user
     * @return Response a response with message `null` if the email address was updated, or a response with a message
     * explaining what went wrong otherwise
     */
    public function set_email(string $uuid, string $email, Mailer $mailer): Response
    {
        // Validate
        if (!filter_var($email, FILTER_VALIDATE_EMAIL) || strlen($email) > self::MAX_EMAIL_LENGTH)
            return new Response(
                payload: ["target" => "email", "message" => "Invalid email address."],
                satisfied: false
            );

        // Begin transaction
        $this->conn->beginTransaction();

        // Check if email address is different
        $stmt = $this->conn->prepare("SELECT email FROM users WHERE uuid=:uuid;");
        $stmt->bindValue(":uuid", $uuid);
        $stmt->execute();
        $email_old = $stmt->fetch(PDO::FETCH_ASSOC)["email"];
        if (hash_equals($email_old, $email))
            return new Response(
                payload: ["target" => "email", "message" => "That is already your email address."],
                satisfied: false
            );

        // TODO: Reject update if email address was changed too recently (within, say, 5 minutes)

        // Check if email address is already in use
        $stmt = $this->conn->prepare("SELECT COUNT(*) AS count
                                      FROM users
                                      WHERE email=:email;");
        $stmt->bindValue(":email", $email);
        $stmt->execute();
        if ($stmt->fetch(PDO::FETCH_ASSOC)["count"] > 0) {
            $this->conn->rollBack();
            return new Response(
                payload: ["target" => "email", "message" => "Email address already in use."],
                satisfied: false
            );
        }

        // Update email address
        $stmt = $this->conn->prepare("UPDATE users
                                      SET email=:email,
                                          email_verification_token=lower(hex(randomblob(16))),
                                          email_verification_token_timestamp=unixepoch()
                                      WHERE uuid=:uuid
                                      RETURNING email_verification_token;");
        $stmt->bindValue(":uuid", $uuid);
        $stmt->bindValue(":email", $email);
        $stmt->execute();
        $email_verification_token = $stmt->fetchAll(PDO::FETCH_ASSOC)[0]["email_verification_token"];

        // Respond
        $this->conn->commit();
        $mailer->queue_verification($email, $email_verification_token);
        return new Response(payload: null, satisfied: true);
    }

    /**
     * Verifies an email address with a token.
     *
     * @param string $email the email address to verify
     * @param string $email_verification_token the token to verify the email address with
     * @return Response the server's response to the request
     */
    public function verify_email(string $email, string $email_verification_token): Response
    {
        if (!filter_var($email, FILTER_VALIDATE_EMAIL) || strlen($email) > self::MAX_EMAIL_LENGTH)
            return new Response(
                payload: ["target" => "email", "message" => "Invalid email address."],
                satisfied: false
            );

        $this->conn->beginTransaction();

        $stmt = $this->conn->prepare("SELECT email_verification_token
                                      FROM users
                                      WHERE email=:email;");
        $stmt->bindValue(":email", $email);
        $stmt->execute();
        $result = $stmt->fetch(PDO::FETCH_ASSOC);

        // TODO: Reformat the following line
        if ($result === false ||
            ($result["email_verification_token"] !== null
                && !hash_equals($result["email_verification_token"], $email_verification_token))) {
            $this->conn->rollBack();
            return new Response(
                payload: ["target" => "email", "message" => "Failed to verify email address."],
                satisfied: false
            );
        }

        if ($result["email_verification_token"] === null) {
            // Email already verified, returning
            $this->conn->rollBack();
            return new Response(payload: null, satisfied: true);
        }

        $stmt = $this->conn->prepare("UPDATE users
                                      SET email_verification_token=null
                                      WHERE email=:email;");
        $stmt->bindValue(":email", $email);
        $stmt->execute();

        $this->conn->commit();
        return new Response(payload: null, satisfied: true);
    }

    /**
     * Resends the email verification email.
     *
     * @param string $uuid the UUID of the user to resend the email for
     * @param Mailer $mailer the mailer to send the email with to the user
     * @return Response a response with message `null` if the email was sent, or a response with a message explaining
     * what went wrong otherwise
     */
    public function resend_verify_email(string $uuid, Mailer $mailer): Response
    {
        $this->conn->beginTransaction();

        $stmt = $this->conn->prepare("SELECT email, email_verification_token, email_verification_token_timestamp
                                      FROM users
                                      WHERE uuid=:uuid;");
        $stmt->bindValue(":uuid", $uuid);
        $stmt->execute();
        $user = $stmt->fetch(PDO::FETCH_ASSOC);

        if (!isset($user["email_verification_token"])) {
            $this->conn->rollBack();
            return new Response(
                payload: ["target" => null, "message" => "Your email address is already verified"],
                satisfied: false
            );
        }

        $verify_email_time = new DateTime("@{$user["email_verification_token_timestamp"]}");
        $minutes_since_last_verify_email = $verify_email_time->diff(new DateTime(), absolute: true)->i;
        if ($minutes_since_last_verify_email < self::MINUTES_BETWEEN_VERIFICATION_EMAILS) {
            $this->conn->rollBack();
            $minutes_left = self::MINUTES_BETWEEN_VERIFICATION_EMAILS - $minutes_since_last_verify_email;
            return new Response(
                payload: [
                    "target" => null,
                    "message" =>
                        "A verification email was sent recently. " .
                        "Please wait $minutes_left more minute(s) before requesting a new email."
                ],
                satisfied: false
            );
        }

        $mailer->queue_verification($user["email"], $user["email_verification_token"]);

        $stmt = $this->conn->prepare("UPDATE users
                                      SET email_verification_token_timestamp=unixepoch()
                                      WHERE uuid=:uuid;");
        $stmt->bindValue(":uuid", $uuid);
        $stmt->execute();

        $this->conn->commit();
        return new Response(payload: null, satisfied: true);
    }

    /**
     * Updates the indicated user's password.
     *
     * @param string $uuid the UUID of the user whose password should be updated
     * @param string $password_old the old password
     * @param string $password_new the new password
     * @param string $password_confirm the confirmation of the new password
     * @return Response a response with message `null` if the password was updated, or a response with a message
     * explaining what went wrong otherwise
     */
    public function set_password(string $uuid, string $password_old, string $password_new,
                                 string $password_confirm): Response
    {
        // Validate
        if (strlen($password_old) > self::MAX_PASSWORD_LENGTH)
            return new Response(
                payload: ["target" => "passwordOld", "message" => "Incorrect old password."],
                satisfied: false
            );
        if (strlen($password_new) < self::MIN_PASSWORD_LENGTH || strlen($password_new) > self::MAX_PASSWORD_LENGTH)
            return new Response(
                payload: [
                    "target" => "passwordNew",
                    "message" => "Your password should be at least 8 and at most 64 characters long."
                ],
                satisfied: false
            );
        if ($password_new !== $password_confirm)
            return new Response(
                payload: ["target" => "passwordConfirm", "message" => "New passwords do not match."],
                satisfied: false
            );

        // Begin transaction
        $this->conn->beginTransaction();

        // Validate old password
        $stmt = $this->conn->prepare("SELECT password FROM users WHERE uuid=:uuid;");
        $stmt->bindValue(":uuid", $uuid);
        $stmt->execute();
        $user = $stmt->fetch(PDO::FETCH_ASSOC);
        if (!password_verify($password_old, $user["password"])) {
            $this->conn->rollBack();
            return new Response(
                payload: ["target" => "passwordOld", "message" => "Incorrect old password."],
                satisfied: false
            );
        }

        // Update password
        $stmt = $this->conn->prepare("UPDATE users
                                      SET password=:password, password_update_time=unixepoch()
                                      WHERE uuid=:uuid;");
        $stmt->bindValue(":uuid", $uuid);
        $stmt->bindValue(":password", password_hash($password_new, PASSWORD_DEFAULT));
        $stmt->execute();

        // Respond
        $this->conn->commit();
        return new Response(payload: null, satisfied: true);
    }
}
